#!/usr/bin/python
### 2000/12/19 Locking of the pipes was added!
#   2000/02/04 Added Loggers for controlling a group of logging processes.
#   2004/05/13 Changed standalone Loggers kill window function.
#
from Tkinter import *
import time, sys, os, string, os.path, signal, commands
import tempfile

from tkTimeout import Timeout

def get_time_string():
   return time.strftime('%H:%M:%S',time.gmtime(time.time()))


def check_path(p, mode = 0775):
   if not os.access(p, os.F_OK):
      try:
         os.makedirs(p, mode)
         print "Created output directory <%s> with mode %o" % (p, mode)
      except OSError:
         print "Cannot access and/or create data output directory <%s>" % (p,)
         print "Check Permissions!"
         sys.exit(1)

class LockError(Exception):
   pass

class SystemCommandError(Exception):
   pass


class Tracker(Frame):
   fmt = '%Y/%m/%d %H:%M:%S'

   def __init__(self, master = None, timeout = 10000):
      Frame.__init__(self, master)
      self.config(relief = GROOVE, bd = 2)

      self.ErrorCount = 0
      self.GoodCount = 0
      self.StartTime = time.time()
      self.LastErrorTime = 0
      self.LastGoodTime = 0

      self.SV_GoodCount = StringVar()
      self.SV_GoodCount.set('0')
      self.SV_ErrorCount = StringVar()
      self.SV_ErrorCount.set('0')

      L_StartTime = Label(self, text =
                  time.strftime(Tracker.fmt, time.gmtime(self.StartTime)),
                  width = 18, bg = 'lightgrey')
      self.SV_LastErrorTime = StringVar()
      self.SV_LastErrorTime.set('')
      self.SV_LastGoodTime = StringVar()
      self.SV_LastGoodTime.set('')

      self.L_GoodCount = Label(self, textvariable = self.SV_GoodCount,
                                 anchor = E, width = 5, bg = 'white')
      self.L_ErrorCount = Label(self, textvariable = self.SV_ErrorCount,
                                 anchor = E, width = 5, bg = 'white')
      self.L_LastGoodTime = Label(self, textvariable = self.SV_LastGoodTime,
                                 anchor = W, width = 18, bg = 'white')
      self.L_LastErrorTime = Label(self, textvariable = self.SV_LastErrorTime,
                                 anchor = W, width = 18, bg = 'white')

      Label(self, text = 'Errors:').grid(row = 3, column = 1, sticky = E)
      self.L_ErrorCount.grid(row = 3, column = 2, sticky = W)
      self.L_LastErrorTime.grid(row = 3, column = 3, sticky = W, padx = 20)

      Label(self, text = 'Good:').grid(row = 2, column = 1, sticky = E)
      self.L_GoodCount.grid(row = 2, column = 2, sticky = W)
      self.L_LastGoodTime.grid(row = 2, column = 3, sticky = W, padx = 20)

      Label(self, text = 'Start:').grid(row = 1, column = 1, sticky = E)
      L_StartTime.grid(row = 1, column = 3, sticky = W, padx = 20)
      self.IV_timed_out = IntVar()   # E.g., for a checkbutton indicator.
      self.timeout = Timeout(master = self, timeout = timeout,
                                    callback = self.timeout_warning)
      self.timeout_warning() # Start with red; turn green when data arrives.


   def update(self, ok):
      if ok:
         self.GoodCount = self.GoodCount + 1
         self.LastGoodTime = time.time()
         self.SV_GoodCount.set('%d' % self.GoodCount)
         self.SV_LastGoodTime.set(
                  time.strftime(Tracker.fmt, time.gmtime(self.LastGoodTime)))
         self.L_LastGoodTime.config(bg = 'white')
         self.timeout.start()
         self.IV_timed_out.set(0)
      else:
         self.ErrorCount = self.ErrorCount + 1
         self.LastErrorTime = time.time()
         self.SV_ErrorCount.set('%d' % self.ErrorCount)
         self.SV_LastErrorTime.set(
                  time.strftime(Tracker.fmt, time.gmtime(self.LastErrorTime)))

   def write(self, fp):
      fp.write('Error statistics at ' +
                  time.strftime(Tracker.fmt, time.gmtime(time.time())) + '\n')
      fp.write('Start time: ' +
                  time.strftime(Tracker.fmt, time.gmtime(self.StartTime)) + '\n')
      fp.write('Good Count: %d     Error Count: %d\n' % (self.GoodCount, self.ErrorCount) )

   def timeout_warning(self):
      self.L_LastGoodTime.config(bg = 'red')
      self.IV_timed_out.set(1)


class LogProcess:  # mainly a data structure for ser_adcp parameters
   def __init__(self, device = 'ttyS0', baud = 9600, program = 'ser_asc',
                  options = '', directory = './',
                  instrument = 'generic', yearbase = 2001):
      # device might end up as a path, e.g. '/dev/ttyS0', after
      # modifications of ser_asc and ser_bin.
      self.device_path = device # In case we might need the original.
      # For making filenames we eliminate the '/':
      self.device = string.replace(device, '/', '-')
      self.baud = baud
      self.program = program
      self.options = options
      self.directory = directory
      self.instrument = instrument
      self.yearbase = yearbase

      pipe_directory = '/tmp/SerialLogger'

      try:
         os.mkdir(pipe_directory)
      except OSError, msg:
         pass # Could check for msg other than [Errno 17] file exists,
              # but don't expect it.

      self.in_pipe = os.path.join(pipe_directory, 'inpipe.%s' % self.device)
      self.out_pipe = os.path.join(pipe_directory, 'outpipe.%s' % self.device)
      self.lock_file = '/var/lock/LCK..%s' % self.device
      self.pipe_lock_file = os.path.join(pipe_directory,
                                          'pipe_pid.%s' % self.device)

      try:
         os.mkfifo(self.in_pipe)
      except OSError, msg:
         pass

      try:
         os.mkfifo(self.out_pipe)
      except OSError, msg:
         pass


   def new_connection(self, start_time = None):
      temp_stderr = tempfile.mktemp()
      if start_time:
         T_opt = " -T %d " % start_time
      else:
         T_opt = ""
      redirection = "2> %s" % temp_stderr     # Bash
      command = "%s -y %d -P %s -b %d -d %s -i %s -o %s %s %s %s &" %  (self.program,
               self.yearbase,
               self.device_path, self.baud, self.directory,
               self.in_pipe, self.out_pipe, T_opt, self.options,
               redirection)
      print command
      ret = os.system(command)
      if ret != 0:
         raise SystemCommandError, "Error in new_connection:\n" \
                                     + command \
                                     + "\n returned %d\n" % ret
      for i in range(10):
         time.sleep(0.1)
         if self.is_running():
            os.remove(temp_stderr)
            return
      msg1 = 'Error in new_connection: %s failed to start.\n' % self.program
      msg2 = 'Command: %s\n' % command
      msg3 = file(temp_stderr).read()
      os.remove(temp_stderr)
      raise SystemCommandError, msg1 + msg2 + msg3

   def is_running(self):
      try:
         lockfile = open(self.lock_file, 'r')
         pid = int(string.split(lockfile.readline())[0])
         lockfile.close()
         if pid == os.getpid():
            return 0 # The lock is held by this process, not by a ser_* process.
         psout = commands.getoutput('ps --noheader -p %d -o pid' % pid)
         if int(psout) == pid:
            return 1
      except (OSError, IOError, ValueError), msg:
         pass
      except Exception, e:
         print e
      return 0


class LogControl:  # mixin for Logger and Loggers
   def make_FB(self, master = None, quit_button = 1):
      """ Frame with horizontal row of start/stop buttons. """
      if not master: master = self
      # Frame with Buttons
      self.FB = Frame(master)
      # Note: it is not packed here because we don't know
      # where we will want to put it.
      # Button Start
      self.BS = Button(self.FB, text = 'Start Logging',
                                 command = self.start_logging)
      self.BS.pack(side = LEFT, expand = YES)
      if quit_button:
         self.BQ = Button(self.FB, text = 'Quit Monitor',
                                    command = self.close_monitor)
         self.BQ.pack(side = LEFT, expand = YES)
      # Button Exit (poor mnemonic and name; fix later)
      self.BE = Button(self.FB, text = 'Stop Logging',
                                 command = self.check_stop_logging)
      self.BE.pack(side = LEFT, expand = YES)
      self.BE.configure(state = DISABLED)

   def confirm_stop(self):
      from SimpleDialog import SimpleDialog
      d = SimpleDialog(self,
                       text = "You can quit this window "
                              "and even log out without "
                              "stopping the underlying data "
                              "acquisition process.\n"
                              "Do you want to stop data acquisition?",
                       buttons = ['Yes', 'No'],
                       default = 0,
                       cancel = 1,
                       title = 'Stop logging?')
      return (d.go() == 0)  # True if Stop Logging is confirmed

   def confirm_close(self):
      from SimpleDialog import SimpleDialog
      d = SimpleDialog(self,
                       text = "The logging processes are still running. "
                              "Do you want to close this monitoring window?",
                       buttons = ['Yes', 'No'],
                       default = 0,
                       cancel = 1,
                       title = 'Close window?')
      return (d.go() == 0)  # True if Close window is confirmed


class Loggers(Frame, LogControl):
   def __init__(self, master = None, yearbase = 2001,
                     quit_button = 1,
                     bindir = '/usr/local/bin'):
      self.self_controlled = 0
      if master == None:
         master = Tk()
         master.title("Logger")
         master.iconname("Log")
         master.protocol("WM_DELETE_WINDOW", self.close_monitor)
         self.self_controlled = 1
      Frame.__init__(self, master)
      self.pack()
      self.yearbase = yearbase
      self.bindir = bindir
      self.LogWins = []
      #self.LSF = None     # LoggersStatusFrame
      self.is_logging = 0
      if self.self_controlled:
         self.make_FB(quit_button = quit_button)
         self.FB.pack(side = TOP, expand = YES)
      self.bind('<Destroy>', self.close_monitor)

   def add(self,  device = 'ttyS0', baud = 9600, program = 'ser_asc',
                  options = '', directory = './',
                  instrument = 'generic',
                  timeout = 1000):
      check_path(directory)
      program = os.path.join(self.bindir, program)

      lp = LogProcess(device = device, baud = baud,
                  program = program, options = options,
                  directory = directory, instrument = instrument,
                  yearbase = self.yearbase)
      self.LogWins.append(Logger(self, lp, control = 0, timeout = timeout))

   def add_list(self, sensors, common_opts, base_dir, timeout):
      optlist = ['device', 'baud', 'program', 'options', 'directory',
                 'instrument', 'timeout']
      kwargs = {}
      for sensor in sensors:
         if sensor['format'] == 'ascii':
            sensor['program'] = 'ser_asc'
         else:
            sensor['program'] = 'ser_bin'
         opts = string.join((common_opts,
                             '-e %s' % sensor['ext'],
                             sensor['opt']))
         if sensor['format'] == 'ascii':
            for s in sensor['strings']:
               opts = " %s '%s'" % (opts, s)
         sensor['options'] = opts
         sensor.setdefault('timeout', timeout)
         sensor['directory'] = os.path.join(base_dir, sensor['subdir'])
         for o in optlist:
            kwargs[o] = sensor[o]
         apply(self.add, (), kwargs)

   def check_stop_logging(self):
      if not self.is_logging:
         return 0
      if self.confirm_stop():
         self.is_logging = 0
         map(Logger.stop_logging, self.LogWins)
         for ii in range(30):
            time.sleep(0.1)
            nrun = self.list_running()[0]
            if nrun == 0:
               break
         if hasattr(self, 'BE'):
            self.BE.configure(state = DISABLED)
            self.BS.configure(state = NORMAL)
         #if self.LSF:
         #   self.LSF.check()
         return 0
      return 1 # Still logging; changed mind

   def close_monitor(self, event = ''):
      if not self.LogWins:
         return
      if self.self_controlled:
         if self.check_stop_logging(): # still logging
            if not self.confirm_close():
               return # jump out
      map(Logger.close_monitor, self.LogWins)
      self.LogWins = []
      if event == '':
         self.unbind('<Destroy>')
      if self.self_controlled:
         self.master.destroy()
      else:
         self.destroy()

   def start_logging(self, start_time = None):
      print "Entering Loggers.start_logging"
      if self.is_logging:
         return
      self.is_logging = 1
      map(Logger.start_logging, self.LogWins, [start_time] * len(self.LogWins))
      if hasattr(self, 'BE'):
         self.BS.configure(state = DISABLED)
         self.BE.configure(state = NORMAL)
      print "Leaving Loggers.start_logging"

   def list_running(self):
      running = []
      for log in self.LogWins:
         if log.logproc.is_running():
            running.append(1)
         else:
            running.append(0)
      nrunning = reduce(lambda x, y: x+y, running)
      return nrunning, running

   def resume(self):
      ''' If all LogWins correspond to active logging
      processes, then resume logging; if only some are
      active, kill them, so we get a fresh start.
      Return 1 if resuming, 0 if starting afresh.
      '''
      print "Entering Loggers.resume"
      nrunning, running = self.list_running()
      if nrunning == 0:             # Nothing to resume.
         return 0
      if nrunning == len(running):  # Resume all.
         self.start_logging()
         return 1
      else:                         # Stop the running ones.
         print "In Loggers.resume, nrunning = %d" % nrunning
         for ii in range(len(running)):
            if running[ii]:
               self.LogWins[ii].start_logging() # to reconnect
               self.LogWins[ii].stop_logging()
         return 0

   def start(self):
      self.start_logging()
      self.mainloop()


class LoggersStatusFrame(Frame):
   def __init__(self, master = None, logs = None):
      Frame.__init__(self, master)
      self.logs = logs # Loggers object
      self.Checkbuttons = []
      for lw in self.logs.LogWins:
         name = lw.logproc.instrument
         IV = lw.stats.IV_timed_out
         cb = Checkbutton(self, variable = IV,
                          text = name,
                          state = DISABLED,
                          indicatoron = 0,
                          selectcolor = 'red', # timed out
                          background = 'grey', # initial: not logging
                          disabledforeground = 'black')
         cb.pack(side = LEFT, expand = YES, fill = BOTH)
         self.Checkbuttons.append(cb)
      self.check()

   def on(self):
      for cb in self.Checkbuttons:
         cb.config(background = 'green')

   def off(self):
      for cb in self.Checkbuttons:
         cb.config(background = 'grey')

   def check(self):
      for ii in range(len(self.Checkbuttons)):
         cb = self.Checkbuttons[ii]
         if self.logs.LogWins[ii].is_logging:
            cb.config(background = 'green')
         else:
            cb.config(background = 'grey')


class Logger(Frame, LogControl):
   def __init__(self, master, logproc, control = 1,
                  timeout = 1000):
      Frame.__init__(self, master)
      self.pack(expand = YES, fill = BOTH)
      self.config(relief = GROOVE, bd = 3)

      self.logproc = logproc    #LogProcess object

      self.nlines = 4  # Can get this in other ways later.

      self.SV_Status = StringVar()
      self.SV_Status.set('Inactive')
      self.is_logging = 0

      self.StatusLine().pack(side = LEFT, expand = NO, pady = 1)
      self.stats = Tracker(self, timeout = timeout)
      self.stats.pack(padx = 5, pady = 5, side = LEFT)

      self.control = control
      if control:
         self.make_FB()

      self.pingline = StringVar()
      self.lines = [''] * self.nlines

      self.MP = Label(self, textvariable = self.pingline,
                           justify = LEFT,
                           anchor = W, height = self.nlines)
      self.MP.config(background = 'lightgreen', width = 80)
      self.MP.config(relief = RIDGE, bd = 2)
      self.MP.pack(side = LEFT, expand = YES, fill = X)

      if control:
         self.bind('<Destroy>', self.close_monitor)
         self.start_logging()

   def StatusLine(self):
      SLframe = Frame(self)
      SLframe.config(relief = RAISED, bd = 2)
      DEV = Label(SLframe, text = '%s\n%s' %
                           (self.logproc.instrument, self.logproc.device),
                           width = 15)
      DEV.pack(side = TOP, padx = 1, pady = 1)
      self.MS = Message(SLframe, textvariable = self.SV_Status)
      self.MS.configure(relief = 'ridge', width = 500, bg = 'white')
      self.MS.pack(side = TOP, pady = 1, padx = 1)
      return SLframe

   def check_stop_logging(self, check = 1):
      if not self.is_logging: return
      if not check or self.confirm_stop():
         self.stop_logging()


   def stop_logging(self):
      if not self.is_logging:   return
      self.tk.deletefilehandler(self.fp)
      self.fp.close()
      cp = open(self.commandpipename, 'w')
      cp.write('X')
      cp.close()
      self.SV_Status.set('Not Logging')
      self.MS.configure(fg = 'red')
      self.is_logging = 0

   def close_monitor(self, event = ''):
      #logfile = open('lograw.log', 'a')
      #self.stats.write(logfile)
      #logfile.close()
      if self.is_logging:
         self.tk.deletefilehandler(self.fp)
         self.fp.close()
      try:
         os.remove(self.logproc.pipe_lock_file)
      except OSError:
         pass
      if self.control and event == '':
         self.unbind('<Destroy>')
      self.destroy()


   def start_logging(self, start_time = None):
      if self.is_logging: return  # so multiple button presses do nothing
      try:
         if not self.logproc.is_running():
            self.logproc.new_connection(start_time)
         self.connect_pipe(self.logproc.out_pipe,
                           self.logproc.in_pipe,
                           self.logproc.pipe_lock_file)
         self.MS.configure(fg = 'green')
      except Exception, msg:
         print msg

   def pipes_are_locked(self, lockfilename):
      try:
         lockfile = open(lockfilename, 'r')
         pid = string.atoi(lockfile.readline())
         psout = commands.getoutput('ps --noheader -p %d -o pid' % pid)
         if string.atoi(psout) == pid:
            return pid
      except (OSError, IOError, ValueError), msg:
         pass # print msg
      return 0

   def lock_pipe(self, lockfilename):
      locking_pid = self.pipes_are_locked(lockfilename)
      if locking_pid != 0 and locking_pid != os.getpid():
         raise LockError(locking_pid)
      try:
         lockfile = open(lockfilename, 'w')
         lockfile.write(`os.getpid()`)
         lockfile.close()
      except (OSError, IOError), msg:
         raise LockError('opening or writing lockfile')


   def connect_pipe(self, datapipename, commandpipename, lockfile):
      self.lock_pipe(lockfile)
      self.datapipename = datapipename
      self.commandpipename = commandpipename
      fd = os.open(self.datapipename, os.O_RDONLY| os.O_NONBLOCK)
      self.fp = os.fdopen(fd)
      self.tk.createfilehandler(self.fp, READABLE, self.pipe_reader)
      self.SV_Status.set('Logging')
      self.is_logging = 1

   def pipe_reader(self, fp, mask):
      line = string.strip(fp.readline())
      # print len(line)
      self.lines = self.lines[-(self.nlines - 1):] + [line]
      self.pingline.set(string.join(self.lines, '\n'))
      # Lines are separated by \n, but there is no trailing \n;
      # so we don't have a blank line at the bottom of the message window.

      if string.find(string.lower(line), 'error') > -1:
         self.MP.config(background = 'pink')
         ok = 0
      else:
         self.MP.config(background = 'lightgreen')
         ok = 1
      self.stats.update(ok)


