#!/usr/bin/python

'''
2001/04/03 EF
Serial port routines.

2002/08/22 EF
Tk-based terminal routines.  Subclass the Tk_terminal class to
get special-purpose terminals.

'''

import termios, sys, os, tty, time, struct, types, select
import string, signal
from Tkinter import *
from threading import *
import Pmw
import Queue
import tkFileDialog, tkMessageBox
from serialport import *

print "tk_terminal ", time.strftime('%H:%M:%S')

class flag_display:
   def __init__(self, parent, port):
      self.port = port
      self.flag_value_dict = self.IntVars()
      self.read_flags()
      self.dialog = Pmw.Dialog(parent,
            buttons = ('Apply', 'Reread', 'Quit'),
            defaultbutton = 'Cancel',
            title = 'Serial port flags',
            command = self.execute)
      #Label(self.dialog.interior(), text = 'Testing').pack()
      self.checkboxes = Pmw.RadioSelect(self.dialog.interior(),
            buttontype = 'checkbutton', orient = VERTICAL)
      self.make_checks()
      self.checkboxes.pack()
      self.dialog.activate()

   def IntVars(self):
      sv = {}
      for key in port_flags.keys():
         sv[key] = IntVar()
      return sv

   def read_flags(self):
      text, self.keys, self.values = self.port.read_flags()
      for i in range(len(self.keys)):
         self.flag_value_dict[self.keys[i]].set(self.values[i])

   def set_flags(self):
      for i in range(len(self.keys)):
         key = self.keys[i]
         new_value = self.flag_value_dict[key].get()
         if new_value != self.values[i]:
            self.port.set_flag(key, new_value)

   def make_checks(self):
      for i in range(len(self.keys)):
         key = self.keys[i]
         self.checkboxes.add(key,
               var = self.flag_value_dict[key])


   def execute(self, result):
      button_function = {'Apply': self.set_flags,
                         'Reread': self.read_flags,
                         'Quit': self.dialog.destroy}
      button_function[result]()

class Timeout(Exception):
   pass

def no_action(event):
   return "break"



class terminal_display(Frame):
   def __init__(self, master):
      Frame.__init__(self, master)
      self.Text = Text(self, width = 80, height = 30, wrap = CHAR,
                       setgrid = NO)
      # setting state = 'disabled' disables everything, not
      # only keyboard input
      s = Scrollbar(self, command = self.Text.yview)
      self.Text.configure(yscrollcommand = s.set)
      self.Text.pack(side = LEFT, expand = YES, fill = BOTH)
      s.pack(side = RIGHT, fill = Y)
      self.Text.bind("<KeyPress>", no_action)

   def append(self, str):
      self.Text.insert(END, str)
      self.Text.see(END)

   def get(self, start = '1.0', end = END):
      return self.Text.get(start, end)

   def mark(self, index = "end - 1 c", gravity = "left",
            name = "BeforeCommand"):
      self.Text.mark_set(name, index)
      self.Text.mark_gravity(name, gravity)

   def unmark(self, name = "BeforeCommand"):
      self.Text.mark_unset(name)

   def get_after_mark(self, name = "BeforeCommand"):
      return self.Text.get(name, END)

   def get_line_with(self, str):
      try:
         ind = self.Text.search(str, END, backwards = 1)
         
         if ind == '':
            return ind
            
         txstring = '%s' % ind
         ls =  '%s.0'   % txstring.split('.')[0]
         le =  '%s.end' % txstring.split('.')[0]

         line = self.Text.get(ls, le)
#         line = self.Text.get(ind + ' linestart', ind + ' lineend')
         return line
      except:
         print "Error in get_line_with"
         return ""

   def get_lines_from(self, str):
      try:
         ind = self.Text.search(str, END, backwards = 1)
         if ind == '':
            return ind
         txstring = '%s' % ind
         ls =  '%s.0'   % txstring.split('.')[0]
         line = self.Text.get(ls, END)
         return line
      except:
         print "Error in get_lines_from"
         return ""


   def waitfor(self, str, timeout = 5):
      # timeout is in seconds, resolved to a tenth
      nloops = max(2, int(timeout * 10))
      for ii in range(0, nloops):
         time.sleep(0.1)
         self.update()
         reply = self.get_after_mark()
         ind = reply.rfind(str)
         if ind != -1:
            self.mark('BeforeCommand + %d chars' % ((ind + len(str)), ))        # Move the mark.
            return reply, ind
         else:
            n = len(reply) - len(str)
            if n > 0:
               self.mark('BeforeCommand + %d chars' % n)
      raise Timeout

   def clear(self):
      self.Text.delete('1.0', END)


class EntryHistory(Entry):
   def __init__(self, master, port, **kw):
      self.send = kw.pop('send_function')
      apply(Entry.__init__, (self, master), kw)
      self.port = port  # pass in whole serial port object,
      self.lines = [''] # so that self.port.stream will be current
      self.nlast = 0
      self.curline = 0
      self.sv = StringVar()
      self.configure(textvariable = self.sv)
      self.bind('<Key-Return>', self.input)
      self.bind('<Key-Up>', self.back)
      self.bind('<Key-Down>', self.forward)

   def input(self, event = ''):
      cl = self.get()
      self.send(cl)
      self.lines.append(cl)
      self.nlast = self.nlast + 1
      self.sv.set('')
      self.curline = 0

   def back(self, event = ''):
      if self.nlast:
         self.curline = (self.curline - 1) % (self.nlast + 1)
         self.sv.set(self.lines[self.curline])

   def forward(self, event = ''):
      if self.nlast:
         self.curline = (self.curline + 1) % (self.nlast + 1)
         self.sv.set(self.lines[self.curline])



class Tk_terminal(serial_port):
   def __init__(self, master = None, device = '/dev/ttyS0',
                      baud = 9600, standalone = 0,
                      termination = '\r\n'):
      serial_port.__init__(self, device = device,
                                 baud = baud,
                                 mode = 'r+');
      if not master:
         master = Tk()
         master.protocol("WM_DELETE_WINDOW", self.close_terminal)
         standalone = 1
      self.master = master
      self.standalone = standalone
      self.termination = termination
      self.Frame = Frame(master, relief = RIDGE, borderwidth = 2)
      self.Frame.pack(expand = YES, fill = BOTH)
      lw = Pmw.LabeledWidget(self.Frame, labelpos = 'w',
                             label_text = 'Transmit line:')
      lw.pack(side = BOTTOM, expand = NO, fill = X)
      self.entry = EntryHistory(lw.interior(), self,
                              send_function = self.send,
                              width = 40,
                              state = DISABLED)
      self.entry.pack(side = BOTTOM, expand = NO, fill = X)
      self.display = terminal_display(self.Frame)
      self.display.bind('<Destroy>', self.close_terminal)
      # Putting the binding here seems to keep the close_terminal
      # function from trying to configure a menu item that is
      # already gone.  There may be a cleaner way to ensure that
      # close_terminal is executed early in the shutdown process.
      self.waitfor = self.display.waitfor
      self.mark = self.display.mark
      self.save = 0
      self.outfile_name = "term_diary.txt"
      self.outfile = None
      self.new_input = Queue.Queue(0)
      self.listening = 0
      self.connectedIV = IntVar()
      self.connectedIV.set(self.listening)
      self.baudIV = IntVar()
      self.baudIV.set(self.get_baud())
      toprow = Frame(self.Frame)
      toprow.pack(side = TOP, expand = NO, fill = X)
      self.toprow = toprow
      self.make_menu(master = toprow)
      self.menubar.pack(side = LEFT, expand = NO, fill = NONE, pady = 5)
      self.statusSV = StringVar()
      self.statusframe = Frame(self.Frame)
      self.statusframe.pack(side = TOP, anchor = W, expand = NO, fill = X)
      self.statusline = Pmw.LabeledWidget(self.statusframe, ##toprow,
                                          labelpos = 'w',
                                          label_text = 'Status:')
      self.statusline.pack(side = LEFT, anchor = W)
      ## Pack the display last, so everything else gets needed
      ## space, and the display gets what's left.
      self.display.pack(side = BOTTOM, expand = YES, fill = BOTH)
      Label(self.statusline.interior(),
            width = 52,
            relief = SUNKEN,
            textvariable = self.statusSV).pack()
      self.set_status()

   def set_status(self, msg = None):
      if msg is None:
         if self.listening:
            C = 'connected.'
         else:
            C = 'not connected.'
         msg = 'Device %s, at %d Baud, is %s' % (
                self.get_device(), self.get_baud(), C)
      self.statusSV.set(msg)

   def start_listening(self):
      self.open_port()      # Might already be open, but make sure.
      self.entry.config(state = NORMAL)
      self.config_menuitem('File', 'Disconnect', state = NORMAL)
      self.config_menuitem('File', 'Connect*', state = DISABLED)
      self.listening = 1
      self.set_status()      # after changing self.listening
      self.thread1 = Thread(target = self.listen)
      self.thread1.setDaemon(0)   # Safety measure; should be unnecessary.
      self.thread1.start()
      self.connectedIV.set(self.listening)
      self.update()

   def update(self):
      try:
         # change to loop count, so it can't get stuck
         # if there is too large a blast of input?
         while 1:
            cc = self.new_input.get_nowait()
            self.display.append(cc)
      except Queue.Empty:
         pass
      if self.listening:
         self.Frame.after(50, self.update)

   def listen(self):
      ''' Thread target for listening to a serial port.
          We set vmin to 0 so it won't block; the sleep
          then ensures the main thread gets a time slice.
      '''
      if not self.fd:
         return
      self.set_cc(vmin = 0, vtime = 1) # vtime in 0.1 s
      termios.tcflush(self.fd, termios.TCIOFLUSH)
      while self.listening:
         cc = os.read(self.fd, 500)
         if cc == "":
            time.sleep(0.001)
         else:
            self.new_input.put(cc)  # via queue to main thread
            if self.save:
               self.outfile.write(cc)


   # Looks like one must use os.read instead of <fileobject>.read
   # for any value of vmin other than 0.

   def stop_listening(self):
      if not self.listening:
         return
      self.listening = 0
      self.entry.config(state = DISABLED)
      self.config_menuitem('File', 'Disconnect', state = DISABLED)
      self.config_menuitem('File', 'Connect*', state = NORMAL)
      self.set_status()
      print "D at start loop"
      while self.thread1.isAlive():
         print "time.sleep"
         time.sleep(0.1)
#        time.sleep(0.05)
      self.close_port()
      self.connectedIV.set(self.listening)

   def send(self, s):
      if self.termination:
         self.stream.write(s + self.termination)
         self.stream.flush()
         self.waitfor(s)
      else:
         for i in range(len(s)):
            self.stream.write(s[i])
            self.stream.flush()
            self.waitfor(s[i])

   def close_terminal(self, event = ''):
      self.stop_listening()
      if self.save:
         self.end_save()
      if self.standalone:
         sys.exit()

   def clear(self):
      self.display.clear()


   def ask_write_file(self):
      fn = tkFileDialog.asksaveasfilename(initialfile ='term_diary.txt',
                     initialdir = './',
                     filetypes = (('Text', '*.txt'), ('All', '*')),
                     parent = self.Frame,
                     title = 'Save Transcript as File')
      if fn == '': return
      try:
         f = open(fn, 'w')
         f.write(self.display.get())
         f.close()
      except:
         tkMessageBox.showerror(message = "Can't write to file %s" % fn)

   def ask_save_file(self):
      dd, ff = os.path.split(self.outfile_name)
      fn = tkFileDialog.asksaveasfilename(initialfile = ff,
                     initialdir = dd,
                     filetypes = (('Text', '*.txt'), ('All', '*')),
                     parent = self.Frame,
                     title = 'Save all received characters')
      if fn == '': return
      self.outfile_name = fn
      self.begin_save()

   def begin_save(self, filename = None):
      if filename is not None:
         self.outfile_name = filename
      self.end_save()
      try:
         self.outfile = open(self.outfile_name, 'ab')
      except:
         tkMessageBox.showerror(message = "Can't write to file %s" %
                                           self.outfilename)
         self.ask_save_file()
      self.config_menuitem('File', 'Save incoming', state = DISABLED)
      self.config_menuitem('File', 'Stop saving', state = NORMAL)
      self.save = 1

   def end_save(self):
      try:
         self.outfile.close()
      except:
         pass
      self.config_menuitem('File', 'Save incoming', state = NORMAL)
      self.config_menuitem('File', 'Stop saving', state = DISABLED)
      self.save = 0

   def ask_device(self):
      fn = tkFileDialog.askopenfilename(initialfile = self.get_device(),
                     initialdir = '/dev',
                     filetypes = (('tty.USA', 'tty.USA*'), ('ttyC', 'ttyC*'),
                                  ('ttyR', 'ttyR*'), ('tty', 'tty*')),
                     parent = self.Frame,
                     title = 'Select serial port device')
#       print fn
#       if fn == '': return
      if not fn : return
      self.set_device(fn)
      self.set_status()


   def change_baud(self, baud = None):
      self.open_port()
      if baud == None:
         baud = self.baudIV.get()
      else:
         self.baudIV.set(baud)
      self.set_baud(baud)
      self.set_status()


   def view_set(self):
      self.open_port()
      flag_display(self.Frame, self)





   def make_menu(self, master = None):
      if not master: master = self.Frame
      mb = Pmw.MenuBar(master)
      #mb.pack(fill = X)
      mb.addmenu('File', '')
      mb.addmenu('Baud', '')
      mb.addmenu('Port', '')
      mb.addmenu('Command', '')

      mb.addmenuitem('File', 'command', '',
                     label = 'Connect to port',
                     command = self.start_listening)
      mb.addmenuitem('File', 'command', '',
                     label = 'Disconnect',
                     command = self.stop_listening,
                     state = DISABLED)
      mb.addmenuitem('File', 'command', '',
                     label = 'Save incoming',
                     command = self.ask_save_file)
      mb.addmenuitem('File', 'command', '',
                     label = 'Stop saving',
                     command = self.end_save,
                     state = DISABLED)

      mb.addmenuitem('File', 'command', '',
                     label = 'Save previous',
                     command = self.ask_write_file)
      mb.addmenuitem('File', 'command', '',
                     label = 'Clear',
                     command = self.clear)
      if self.standalone:
         mb.addmenuitem('File', 'command', '',
                     label = 'Quit',
                     command = self.close_terminal)


      mb.addmenuitem('Command', 'command', '',
                     label = 'Send Break',
                     command = self.send_break)



      bauds = baud_table.keys()
      bauds.sort()
      for b in bauds:
         mb.addmenuitem('Baud', 'radiobutton', '',
                     label = str(b),
                     var = self.baudIV,
                     command = self.change_baud,
                     value = b)

      mb.addmenuitem('Port', 'command', '',
                     label = 'Device',
                     command = self.ask_device)
      mb.addmenuitem('Port', 'command', '',
                     label = 'View/Set',
                     command = self.view_set)
      self.menubar = mb

   def config_menuitem(self, menu, item, **kw):
      m = self.menubar.component(menu + '-menu')
      apply(m.entryconfigure, (item,), kw)


if __name__ == '__main__':
   device = '/dev/ttyS0'
   baud = 9600
   if len(sys.argv) > 1:
      device = sys.argv[1]
   if len(sys.argv) > 2:
      baud = int(sys.argv[2])
   T = Tk_terminal(device = device, master = Tk(), baud = baud, standalone = 1)
   T.Frame.mainloop()

